
#include "astar.h"
#include <functional>

std::list<Vertex *> findpath(GlobalMap &gm, float currentpos_x, float currentpos_y, Vertex *goal_vtx)
{
gm.count_vtx_tested = 0;
gm.count_are_neighbors = 0;
	std::list<Vertex *> opennodes;
	std::list<Vertex *> closednodes;

	Vertex startvertex;
	int cellx = (int)(currentpos_x / gm.cellSize);
	int celly = (int)(currentpos_y / gm.cellSize);
	startvertex.x = cellx;
	startvertex.y = celly;
	startvertex.cost = 0;
	startvertex.overall_cost = distance(&startvertex, goal_vtx);
	startvertex.parent = NULL;
	for(int i = 0; i < gm.nvertices; i++)
		if (gm.areNeighbors(cellx, celly, gm.vertices[i]->x, gm.vertices[i]->y))
			startvertex.neighbors.push_back(Vertex::NeighborVertex(gm.vertices[i], distance(&startvertex, gm.vertices[i]), true));

	Vertex *current_vtx = &startvertex;
	closednodes.push_back(current_vtx);

	bool no_solution = false;
	while(current_vtx != goal_vtx && !no_solution)
	{
gm.count_vtx_tested++;
//std::cout << current_vtx->x << "," << current_vtx->y <<" ; ";
		for(NeighborList::iterator neighbor = current_vtx->neighbors.begin(); neighbor != current_vtx->neighbors.end(); neighbor++)
		{
			// check if neighbor is in closed list, and zap if so
			bool neighbor_is_in_closed = false;
			for(std::list<Vertex *>::iterator it = closednodes.begin(); it != closednodes.end(); it++)
				if ((*it) == (*neighbor).ptr) { neighbor_is_in_closed = true; break; }
			if (neighbor_is_in_closed) continue; 

			// check if is in open and update if so
			float cost = current_vtx->cost + distance(current_vtx, (*neighbor).ptr);
			float overall_cost = cost + distance((*neighbor).ptr, goal_vtx);

			bool was_inopen = false;
			for(std::list<Vertex *>::iterator it = opennodes.begin(); it != opennodes.end(); it++)
				if ((*it) == (*neighbor).ptr)
				{
					was_inopen = true;
					if (overall_cost < (*it)->overall_cost)
					{
						(*it)->overall_cost = overall_cost;
						(*it)->cost = cost;
						(*it)->parent = current_vtx;
					}
				}
			
			if (!was_inopen)
			{
				(*neighbor).ptr->parent = current_vtx;
				(*neighbor).ptr->cost = cost;
				(*neighbor).ptr->overall_cost = overall_cost;
				opennodes.push_back((*neighbor).ptr);
			}
		}

		// find next vertex (min cost vertex in opennodes)
		if (!opennodes.empty())
		{ 
			std::list<Vertex *>::iterator mincost_iterator = opennodes.begin();
			for(std::list<Vertex *>::iterator it = opennodes.begin(); it != opennodes.end(); it++)
				if ((*it)->overall_cost < (*mincost_iterator)->overall_cost) mincost_iterator = it;

			current_vtx = (*mincost_iterator);
			opennodes.erase(mincost_iterator);
			closednodes.push_back(current_vtx);
		} else no_solution = true;
	}

	std::list<Vertex *> result;
	if (no_solution) return result; else
	while (current_vtx->parent != NULL)
	{
		result.push_front(current_vtx);
		current_vtx = current_vtx->parent;
	}

	return result;


/*

	vertex bestopennode;
	int j;

	for (int i = 0; i < nodelist.size(); i++)
	{
	if (areNeighbors(currentpos_x, currentpos_y, nodelisd[i].x, nodelist[i].y) == true)
	{
	neighbors.push_back();
	}
	for (int i = 0; i < neighbors.size(); i++)
	{
	if (opennodes.contains(neighbors[i])==false)
	{
		if (neighbors[i].distance > opennodes[size-1].distance)
		{
		opennodes.popback();
		opennodes.popback();
		opennodes.pushback(parent(neighbors[i])); //voir comment stocker les parents
		opennodes.pushback(neighbors[i]);
		}
		else
		{
opennodes.pushback(parent(neighbors[i]);
opennodes.pushback(neighbors[i]);
		}
	}
	if (opennodes.size !=0) {
	bestopennode = opennodes[0];
	for (i=1; i<opennodes.size(); i++)
	{
	if (opennodes[i].distance > bestopennode[0].distance)
	{
		bestopennode = opennodes[i];
		j=i;
	}
	closednodes.pushback(bestopennode);
	opennodes.remove(bestopennode);
	}
	}
iterer avec le dernier noeud de la liste fermee come noeud courant;
	}
	}
	rcuprer les parents du noeud d'arrive;

	*/

}

/*
with belonging to opennodes and closednodes stored
*/
std::list<Vertex *> findpath_opt1(GlobalMap &gm, float currentpos_x, float currentpos_y, Vertex *goal_vtx)
{
gm.count_vtx_tested = 0;
gm.count_are_neighbors = 0;
	std::list<Vertex *> opennodes;
	std::list<Vertex *> closednodes;

	Vertex startvertex;
	int cellx = (int)(currentpos_x / gm.cellSize);
	int celly = (int)(currentpos_y / gm.cellSize);
	startvertex.x = cellx;
	startvertex.y = celly;
	startvertex.cost = 0;
	startvertex.overall_cost = distance(&startvertex, goal_vtx);
	startvertex.parent = NULL;
	for(int i = 0; i < gm.nvertices; i++)
		if (gm.areNeighbors(cellx, celly, gm.vertices[i]->x, gm.vertices[i]->y))
			startvertex.neighbors.push_back(Vertex::NeighborVertex(gm.vertices[i], distance(&startvertex, gm.vertices[i]), true));

	Vertex *current_vtx = &startvertex;
	closednodes.push_back(current_vtx); current_vtx->is_in_closed = true;

	for(int i = 0; i < gm.nvertices; i++) gm.vertices[i]->is_in_closed = gm.vertices[i]->is_in_open = false;

	bool no_solution = false;
	while(current_vtx != goal_vtx && !no_solution)
	{
gm.count_vtx_tested++;
//std::cout << current_vtx->x << "," << current_vtx->y <<" ; ";
		for(NeighborList::iterator neighbor = current_vtx->neighbors.begin(); neighbor != current_vtx->neighbors.end(); neighbor++)
		{
			// check if neighbor is in closed list, and zap if so
			if ((*neighbor).ptr->is_in_closed) continue;

			// check if is in open and update if so
			float cost = current_vtx->cost + distance(current_vtx, (*neighbor).ptr);
			float overall_cost = cost + distance((*neighbor).ptr, goal_vtx);

			if ((*neighbor).ptr->is_in_open)
			{
				if (overall_cost < (*neighbor).ptr->overall_cost)
				{
					(*neighbor).ptr->overall_cost = overall_cost;
					(*neighbor).ptr->cost = cost;
					(*neighbor).ptr->parent = current_vtx;
				}			
			} else
			{
				(*neighbor).ptr->parent = current_vtx;
				(*neighbor).ptr->cost = cost;
				(*neighbor).ptr->overall_cost = overall_cost;
				opennodes.push_back((*neighbor).ptr); (*neighbor).ptr->is_in_open = true;
			}
		}

		// find next vertex (min cost vertex in opennodes)
		if (!opennodes.empty())
		{ 
			std::list<Vertex *>::iterator mincost_iterator = opennodes.begin();
			for(std::list<Vertex *>::iterator it = opennodes.begin(); it != opennodes.end(); it++)
				if ((*it)->overall_cost < (*mincost_iterator)->overall_cost) mincost_iterator = it;

			current_vtx = (*mincost_iterator);
			opennodes.erase(mincost_iterator); current_vtx->is_in_open = false;
			closednodes.push_back(current_vtx); current_vtx->is_in_closed = true;
		} else no_solution = true;
	}

	std::list<Vertex *> result;
DEBUG("no_soluton %d", no_solution);
	if (no_solution) return result; else
	while (current_vtx->parent != NULL)
	{
		result.push_front(current_vtx);
		current_vtx = current_vtx->parent;
	}

DEBUG("size %d", result.size());

	return result;




}


struct compareNVp : std::binary_function<Vertex::NeighborVertex *, Vertex::NeighborVertex *, bool>
{
	bool operator()(const Vertex::NeighborVertex *nv1, const Vertex::NeighborVertex *nv2) const 
	{
		return (nv1->ptr->overall_cost > nv2->ptr->overall_cost);
	}
};

/*
with priority queue
*/
std::list<Vertex *> findpath_opt2(GlobalMap &gm, float currentpos_x, float currentpos_y, Vertex *goal_vtx)
{
gm.count_vtx_tested = 0;
gm.count_are_neighbors = 0;
	std::priority_queue<Vertex::NeighborVertex *, std::vector<Vertex::NeighborVertex *>, compareNVp> opennodes;
	std::list<Vertex *> closednodes;

	Vertex startvertex;
	int cellx = (int)(currentpos_x / gm.cellSize);
	int celly = (int)(currentpos_y / gm.cellSize);
	startvertex.x = cellx;
	startvertex.y = celly;
	startvertex.cost = 0;
	startvertex.overall_cost = distance(&startvertex, goal_vtx);
	startvertex.parent = NULL;
	for(int i = 0; i < gm.nvertices; i++)
		if (gm.areNeighbors(cellx, celly, gm.vertices[i]->x, gm.vertices[i]->y))
			startvertex.neighbors.push_back(Vertex::NeighborVertex(gm.vertices[i], distance(&startvertex, gm.vertices[i]), true));

	Vertex *current_vtx = &startvertex;
	closednodes.push_back(current_vtx); current_vtx->is_in_closed = true;

	for(int i = 0; i < gm.nvertices; i++) gm.vertices[i]->is_in_closed = gm.vertices[i]->is_in_open = false;

	bool no_solution = false;
	while(current_vtx != goal_vtx && !no_solution)
	{
gm.count_vtx_tested++;
//std::cout << current_vtx->x << "," << current_vtx->y <<" ; ";
		for(NeighborList::iterator neighbor = current_vtx->neighbors.begin(); neighbor != current_vtx->neighbors.end(); neighbor++)
		{
			// check if neighbor is in closed list, and zap if so
			if ((*neighbor).ptr->is_in_closed) continue;

			// check if is in open and update if so
			float cost = current_vtx->cost + distance(current_vtx, (*neighbor).ptr);
			float overall_cost = cost + distance((*neighbor).ptr, goal_vtx);

			if ((*neighbor).ptr->is_in_open)
			{
				if (overall_cost < (*neighbor).ptr->overall_cost)
				{
					(*neighbor).ptr->parent = current_vtx;
					(*neighbor).ptr->overall_cost = overall_cost;
					(*neighbor).ptr->cost = cost;
					(*neighbor).removed = true;
					opennodes.push(&(*neighbor));
				}			
			} else
			{
				(*neighbor).ptr->parent = current_vtx;
				(*neighbor).ptr->cost = cost;
				(*neighbor).ptr->overall_cost = overall_cost;
				opennodes.push(&(*neighbor)); (*neighbor).ptr->is_in_open = true;
			}
		}
		
		// find next vertex (min cost vertex in opennodes)
		bool found_vertex = false;
		Vertex::NeighborVertex *top = NULL;
		while (!opennodes.empty() && !found_vertex)
		{
			top = opennodes.top();
			if (top->removed) { opennodes.pop(); top->ptr->is_in_open = false; continue; }
			if (!top->tested_neighbor)
				if (!gm.areNeighbors(top->ptr->parent->x, top->ptr->parent->y, top->ptr->x, top->ptr->y))
					{ opennodes.pop(); top->ptr->is_in_open = false; continue; } else 
					top->tested_neighbor = true;
			found_vertex = true;
		}
		if (found_vertex)
		{
			current_vtx = top->ptr;
			opennodes.pop(); current_vtx->is_in_open = false;
			closednodes.push_back(current_vtx); current_vtx->is_in_closed = true;
		} else no_solution = true;
	}

	std::list<Vertex *> result;
	if (no_solution) return result; else
	while (current_vtx->parent != NULL)
	{
		result.push_front(current_vtx);
		current_vtx = current_vtx->parent;
	}

	return result;

}


/*
starting from goal
*/
std::list<Vertex *> findpath_opt3(GlobalMap &gm, float currentpos_x, float currentpos_y, Vertex *goal_vtx)
{
gm.count_vtx_tested = 0;
gm.count_are_neighbors = 0;
	std::priority_queue<Vertex::NeighborVertex *, std::vector<Vertex::NeighborVertex *>, compareNVp> opennodes;
	std::list<Vertex *> closednodes;

	Vertex startvertex;
	int cellx = (int)(currentpos_x / gm.cellSize);
	int celly = (int)(currentpos_y / gm.cellSize);
	startvertex.x = cellx;
	startvertex.y = celly;
	startvertex.is_in_closed = startvertex.is_in_open = false;
//	startvertex.cost = 0;
//	startvertex.overall_cost = distance(&startvertex, goal_vtx);
//	startvertex.parent = NULL;
	for(int i = 0; i < gm.nvertices; i++)
		if (gm.areNeighbors(cellx, celly, gm.vertices[i]->x, gm.vertices[i]->y))
		gm.vertices[i]->neighbors.push_back(Vertex::NeighborVertex(&startvertex, distance(&startvertex, gm.vertices[i]), true));
		//startvertex.neighbors.push_back(Vertex::NeighborVertex(gm.vertices[i], distance(&startvertex, gm.vertices[i]), false));

	for(int i = 0; i < gm.nvertices; i++) gm.vertices[i]->is_in_closed = gm.vertices[i]->is_in_open = false;

	goal_vtx->cost = 0;
	goal_vtx->overall_cost = distance(&startvertex, goal_vtx);
	goal_vtx->parent = NULL;
	Vertex *current_vtx = goal_vtx;
	closednodes.push_back(current_vtx); current_vtx->is_in_closed = true;


	bool no_solution = false;
	while(current_vtx != &startvertex && !no_solution)
	{
gm.count_vtx_tested++;
//std::cout << current_vtx->x << "," << current_vtx->y <<" ; ";
		for(NeighborList::iterator neighbor = current_vtx->neighbors.begin(); neighbor != current_vtx->neighbors.end(); neighbor++)
		{
			// check if neighbor is in closed list, and zap if so
			if ((*neighbor).ptr->is_in_closed) continue;

			// check if is in open and update if so
			float cost = current_vtx->cost + distance(current_vtx, (*neighbor).ptr);
			float overall_cost = cost + distance((*neighbor).ptr, &startvertex);

			if ((*neighbor).ptr->is_in_open)
			{
				if (overall_cost < (*neighbor).ptr->overall_cost)
				{
					(*neighbor).ptr->parent = current_vtx;
					(*neighbor).ptr->overall_cost = overall_cost;
					(*neighbor).ptr->cost = cost;
					(*neighbor).removed = true;
					opennodes.push(&(*neighbor));
				}			
			} else
			{
				(*neighbor).ptr->parent = current_vtx;
				(*neighbor).ptr->cost = cost;
				(*neighbor).ptr->overall_cost = overall_cost;
				opennodes.push(&(*neighbor)); (*neighbor).ptr->is_in_open = true;
			}
		}
		
		// find next vertex (min cost vertex in opennodes)
		bool found_vertex = false;
		Vertex::NeighborVertex *top = NULL;
		while (!opennodes.empty() && !found_vertex)
		{
			top = opennodes.top();
			if (top->removed) { opennodes.pop(); top->ptr->is_in_open = false; continue; }
			if (!top->tested_neighbor)
				if (!gm.areNeighbors(top->ptr->parent->x, top->ptr->parent->y, top->ptr->x, top->ptr->y))
					{ opennodes.pop(); top->ptr->is_in_open = false; continue; } else 
					top->tested_neighbor = true;
			found_vertex = true;
		}
		if (found_vertex)
		{
			current_vtx = top->ptr;
			opennodes.pop(); current_vtx->is_in_open = false;
			closednodes.push_back(current_vtx); current_vtx->is_in_closed = true;
		} else no_solution = true;
	}

	std::list<Vertex *> result;
	if (no_solution || current_vtx->parent == NULL) return result; else
	do
	{
		current_vtx = current_vtx->parent;
		result.push_back(current_vtx);
	} while (current_vtx->parent != NULL);

	for(int i = 0; i < gm.nvertices; i++)
		if (gm.vertices[i]->neighbors.back().ptr ==  &startvertex) gm.vertices[i]->neighbors.pop_back();

	
	return result;

}

/*
with cost computed before neighborhood
*/
std::list<Vertex *> findpath_opt4(GlobalMap &gm, float currentpos_x, float currentpos_y, Vertex *goal_vtx)
{
gm.count_vtx_tested = 0;
gm.count_are_neighbors = 0;
	std::priority_queue<Vertex::NeighborVertex *, std::vector<Vertex::NeighborVertex *>, compareNVp> opennodes;
	std::list<Vertex *> closednodes;

	Vertex startvertex;
	int cellx = (int)(currentpos_x / gm.cellSize);
	int celly = (int)(currentpos_y / gm.cellSize);
	startvertex.x = cellx;
	startvertex.y = celly;
	startvertex.is_in_closed = startvertex.is_in_open = false;
//	startvertex.cost = 0;
//	startvertex.overall_cost = distance(&startvertex, goal_vtx);
//	startvertex.parent = NULL;
	for(int i = 0; i < gm.nvertices; i++)
		//if (gm.areNeighbors(cellx, celly, gm.vertices[i]->x, gm.vertices[i]->y))
		gm.vertices[i]->neighbors.push_back(Vertex::NeighborVertex(&startvertex, distance(&startvertex, gm.vertices[i]), false));
		//startvertex.neighbors.push_back(Vertex::NeighborVertex(gm.vertices[i], distance(&startvertex, gm.vertices[i]), false));

	for(int i = 0; i < gm.nvertices; i++) gm.vertices[i]->is_in_closed = gm.vertices[i]->is_in_open = false;

	goal_vtx->cost = 0;
	goal_vtx->overall_cost = distance(&startvertex, goal_vtx);
	goal_vtx->parent = NULL;
	Vertex *current_vtx = goal_vtx;
	closednodes.push_back(current_vtx); current_vtx->is_in_closed = true;


	bool no_solution = false;
	while(current_vtx != &startvertex && !no_solution)
	{
gm.count_vtx_tested++;
//std::cout << current_vtx->x << "," << current_vtx->y <<" ; ";
		for(NeighborList::iterator neighbor = current_vtx->neighbors.begin(); neighbor != current_vtx->neighbors.end(); neighbor++)
		{
			// check if neighbor is in closed list, and zap if so
			if ((*neighbor).ptr->is_in_closed) continue;

			// check if is in open and update if so
			float cost = current_vtx->cost + distance(current_vtx, (*neighbor).ptr);
			float overall_cost = cost + distance((*neighbor).ptr, &startvertex);

			if ((*neighbor).ptr->is_in_open)
			{
				if (overall_cost < (*neighbor).ptr->overall_cost)
				{
					(*neighbor).ptr->parent = current_vtx;
					(*neighbor).ptr->overall_cost = overall_cost;
					(*neighbor).ptr->cost = cost;
					(*neighbor).removed = true;
					opennodes.push(&(*neighbor));
				}			
			} else
			{
				(*neighbor).ptr->parent = current_vtx;
				(*neighbor).ptr->cost = cost;
				(*neighbor).ptr->overall_cost = overall_cost;
				opennodes.push(&(*neighbor)); (*neighbor).ptr->is_in_open = true;
			}
		}
		
		// find next vertex (min cost vertex in opennodes)
		bool found_vertex = false;
		Vertex::NeighborVertex *top = NULL;
		while (!opennodes.empty() && !found_vertex)
		{
			top = opennodes.top();
			if (top->removed) { opennodes.pop(); top->ptr->is_in_open = false; continue; }
			if (!top->tested_neighbor)
				if (!gm.areNeighbors(top->ptr->parent->x, top->ptr->parent->y, top->ptr->x, top->ptr->y))
					{ opennodes.pop(); top->ptr->is_in_open = false; continue; } else 
					top->tested_neighbor = true;
			found_vertex = true;
		}
		if (found_vertex)
		{
			current_vtx = top->ptr;
			opennodes.pop(); current_vtx->is_in_open = false;
			closednodes.push_back(current_vtx); current_vtx->is_in_closed = true;
		} else no_solution = true;
	}

	std::list<Vertex *> result;
	if (no_solution || current_vtx->parent == NULL) return result; else
	do
	{
		current_vtx = current_vtx->parent;
		result.push_back(current_vtx);
	} while (current_vtx->parent != NULL);

	for(int i = 0; i < gm.nvertices; i++)
		gm.vertices[i]->neighbors.pop_back();

	
	return result;

}







