/********************************************
Personal library of Cyril Roussillon
http://crteknologies.free.fr
cyril42e@gmail.com
2006-2007
*********************************************/

#include "cyril-config.h"

#ifdef _USE_XML_
	#include "cyril-xml.h"
#endif



//////////////////////////////////////////////////////////////////
// config
//////////////////////////////////////////////////////////////////

int argc_ = 0;
char **argv_ = NULL;

const char* CFG_TypeNames[pNType] = { "sep", "bool", "enum", "int", "real", "string", "file" };

int CFG_toolCount = 0;
int CFG_toolMaxCount = 0;
CFG_Tool *CFG_tools = NULL;
int CFG_tool = -1;

//CFG_Param *CFG_params = NULL;
char *CFG_configFile = NULL;
char *CFG_helpFile = NULL;

const char sep = ';';
const char cfg = '@';

/**
args, usage, help
*/

void CFG_setVal(CFG_Param *param, char *arg);
void CFG_printUsage(char *argv0);
void CFG_printHelp(char *argv0, int tool);
#ifdef _USE_XML_
	void CFG_loadOptionsXML(char *xmlname, char *type, int tool);
#endif
void CFG_loadOptionsCLI(int argc, char **argv);


//////////////////////////////////////////////////////////////////
// 
//////////////////////////////////////////////////////////////////

CFG_Tool* CFG_getTool(const char *name)
{
	for(int i = 0; i < CFG_toolCount; i++)
		if (!strcmp(name, CFG_tools[i].name)) return &CFG_tools[i];
	return NULL;
}

bool CFG_toolIs(const char *name)
{
	return (!strcmp(name, CFG_tools[CFG_tool].name));
}


bool CFG_loadArgs(int argc, char **argv)
{
	argc_ = argc;
	argv_ = argv;

	// look for the tool asked
	CFG_tool = -3; // default
	if (argc <= 1) CFG_tool = -3; else
	if (!strcmp(argv[1], "--usage")) CFG_tool = -2; else
	if (!strcmp(argv[1], "--help")) CFG_tool = -1; else
	for(int i = 0; i < CFG_toolCount; i++)
		if (!strncmp(argv[1], "--", 2) && !strcmp(argv[1]+2, CFG_tools[i].name))
			{ CFG_tool = i; break; }
	int offset_params = 2;
	if (CFG_tool == -3)
	{ 
		offset_params = 1; 
		CFG_tool = (CFG_toolCount == 1 ? 0 : -2); 
	}
	
	// some init
	for(int i = 0; i < CFG_toolCount; i++)
	{
		CFG_tools[i].params[0] = (CFG_Param){&CFG_configFile, NULL, pFile, 1,
			"config", "The name and path of the xml configuration file", 0, "", ""};
		char configname[200]; sprintf(configname, "config-%s.xml", CFG_tools[i].name);
		CFG_setVal(CFG_tools[i].params+0, configname);
		CFG_tools[i].params[1] = (CFG_Param){&CFG_helpFile, NULL, pFile, 1,
			"help", "The name and path of the xml help file", 0, "", ""};
		CFG_setVal(CFG_tools[i].params+1, "help.xml");
	}
	
	// and loading of config and help params
	int tool = (CFG_tool >= 0 ? CFG_tool : 0);
	for(int j = 0; j < argc; j++)
	for(int i = 0; i < 2; i++)
	{
		int optlen = strlen(CFG_tools[tool].params[i].name); 
		if (argv[j][0] == '-' && !strncmp(argv[j]+1, CFG_tools[tool].params[i].name, optlen) && argv[j][optlen+1] == '=')
		{ 
			CFG_tools[tool].params[i].val = 2;
			CFG_setVal(CFG_tools[tool].params+i, argv[j]+optlen+2); 
			break; 
		}
	}
	
	// parse help commands and check consistency
	if (CFG_tool == -2)
	{
		CFG_printUsage(argv[0]);
		return false;
	}
	if (CFG_tool == -1)
	{
		int tool = -1;
		if (argc == 3)
			for(int i = 0; i < CFG_toolCount; i++)
				if (!strcmp(argv[2], CFG_tools[i].name))
					{ tool = i; break; }
		if (argc == 2 || tool != -1) CFG_printHelp(argv[0], tool);
		else CFG_printUsage(argv[0]);
		return false; 
	}

	
	// parse xml options
	#ifdef _USE_XML_
	if (CFG_tools[CFG_tool].params[0].str != "")
		CFG_loadOptionsXML(CFG_tools[CFG_tool].params[0].str, CFG_tools[CFG_tool].configType, CFG_tool);
	#endif
	
	//parse cli options
	CFG_loadOptionsCLI(argc-offset_params, argv+offset_params);

	return true;
}


bool CFG_initOtherTool(const char *tool, char *config)
{
	// look for the tool asked
	int cfg_tool = -3; // default
	for(int i = 0; i < CFG_toolCount; i++)
		if (!strcmp(tool, CFG_tools[i].name))
			{ cfg_tool = i; break; }
	if (cfg_tool == -3) return false;
	
	CFG_setVal(CFG_tools[cfg_tool].params+0, config);

	// parse xml options
	#ifdef _USE_XML_
	if (CFG_tools[cfg_tool].params[0].str != "")
		CFG_loadOptionsXML(CFG_tools[cfg_tool].params[0].str, CFG_tools[cfg_tool].configType, cfg_tool);
	#endif
	
	return true;
}

void CFG_printUsage(char *argv0)
{
	// help tools
	fprintf(stderr, "\n" clBYELLOW "Usage:" clNORMAL "\n");
	fprintf(stderr, "  " clBWHITE "%s --usage\n" clNORMAL, argv0);
	fprintf(stderr, "  " clBWHITE "%s --help [", argv0);
	for(int i = 0; i < CFG_toolCount; i++) 
		fprintf(stderr, (i == 0 ? "%s" : " | %s"), CFG_tools[i].name);
	fprintf(stderr, "]" clNORMAL "\n");
	fprintf(stderr, "\n");
	// default tool
	fprintf(stderr, "  " clBWHITE "%s", argv0);
	#ifdef _USE_XML_
	fprintf(stderr, " [-config=name.xml] [-help=name.xml]");
	#endif
	fprintf(stderr, " [-option1=val1, ...]" clNORMAL " ('%s' tool by default)\n", CFG_tools[0].name);
	// all tools
	for(int i = 0; i < CFG_toolCount; i++)
	{
		fprintf(stderr, "  " clBWHITE "%s --%s", argv0, CFG_tools[i].name);
		#ifdef _USE_XML_
		fprintf(stderr, " [-config=name.xml] [-help=name.xml]");
		#endif
		fprintf(stderr, " [-option1=val1, ...]" clNORMAL "\n");
	}
	fprintf(stderr, "\n");
}
	
void CFG_printToolHelp(char *argv0, int tool)
{
	// help tools
	fprintf(stdout, "\n" clBGREEN "# TOOL %s" clNORMAL "\n", CFG_tools[tool].name);
	
	// usage
	fprintf(stdout, clBYELLOW "\n  Usage:" clNORMAL "\n");
	if (tool == 0)
	{
		fprintf(stdout, "    " clBWHITE "%s", argv0);
		#ifdef _USE_XML_
		fprintf(stdout, " [-config=name.xml] [-help=name.xml]");
		#endif
		fprintf(stdout, " [-option1=val1, ...]" clNORMAL " ('%s' tool by default)\n", CFG_tools[tool].name);
	}
	fprintf(stdout, "    " clBWHITE "%s --%s", argv0, CFG_tools[tool].name);
	#ifdef _USE_XML_
	fprintf(stdout, " [-config=name.xml] [-help=name.xml]");
	#endif
	fprintf(stdout, " [-option1=val1, ...]" clNORMAL "\n");
	
	// description
	fprintf(stdout, clBYELLOW "\n  Description:" clNORMAL "\n");
	fprintf(stdout, "    %s\n", CFG_tools[tool].desc);

	// params
	fprintf(stdout, clBYELLOW "\n  Parameters:" clNORMAL "\n");
	int optlen, rangelen, deflen, typelen;
	int maxoptlen = 0, maxrangelen = 0, maxdeflen = 0;
	for(int i = 0; i < 2; i++) // calc max optlen
	{ 
		if (CFG_tools[tool].params[i].type == pSep) continue;
		optlen = strlen(CFG_tools[tool].params[i].name); 
		if (optlen > maxoptlen) maxoptlen = optlen; 
		rangelen = strlen(CFG_tools[tool].params[i].range); 
		if (rangelen > maxrangelen) maxrangelen = rangelen; 
	}
	for(int i = 0; i < CFG_tools[tool].optionCount; i++) // calc max optlen
	{ 
		if (CFG_tools[tool].options[i].type == pSep) continue;
		optlen = strlen(CFG_tools[tool].options[i].name); 
		if (optlen > maxoptlen) maxoptlen = optlen; 
		rangelen = strlen(CFG_tools[tool].options[i].range); 
		if (rangelen > maxrangelen) maxrangelen = rangelen; 
		deflen = strlen(CFG_tools[tool].options[i].default_val); 
		if (deflen > maxdeflen) maxdeflen = deflen; 
	}
	for(int i = 0; i < 2; i++)
	{
		if (CFG_tools[tool].params[i].type == pSep) 
			fprintf(stdout, "\n");
		else
		{
			optlen = strlen(CFG_tools[tool].params[i].name);
			rangelen = strlen(CFG_tools[tool].params[i].range); 
			typelen = strlen(CFG_TypeNames[CFG_tools[tool].params[i].type]);
			fprintf(stdout, "    " clBWHITE "%s%s" clNORMAL, CFG_tools[tool].params[i].name, FILL_BLANK(maxoptlen-optlen));
			if (CFG_tools[tool].params[i].size == 1)
				fprintf(stdout, "  %s%s", CFG_TypeNames[CFG_tools[tool].params[i].type], FILL_BLANK(8-typelen)); else
				fprintf(stdout, "  %s^%d%s", CFG_TypeNames[CFG_tools[tool].params[i].type], CFG_tools[tool].params[i].size, FILL_BLANK(8-typelen-2));
			fprintf(stdout, " %s  %s",
				FILL_BLANK(maxdeflen), 
				CFG_tools[tool].params[i].desc);
			if (CFG_tools[tool].params[i].range == NULL || CFG_tools[tool].params[i].range[0] == 0) 
				fprintf(stdout, "\n"); else
				fprintf(stdout, " : %s\n", CFG_tools[tool].params[i].range);
		}
	}
	
	// options
	fprintf(stdout, clBYELLOW "\n  Options:" clNORMAL "\n");
	for(int i = 0; i < CFG_tools[tool].optionCount; i++)
	{
		if (CFG_tools[tool].options[i].type == pSep) 
			fprintf(stdout, "\n");
		else
		{
			optlen = strlen(CFG_tools[tool].options[i].name);
			rangelen = strlen(CFG_tools[tool].options[i].range); 
			deflen = strlen(CFG_tools[tool].options[i].default_val); 
			typelen = strlen(CFG_TypeNames[CFG_tools[tool].options[i].type]);
			fprintf(stdout, "    " clBWHITE "%s%s" clNORMAL, CFG_tools[tool].options[i].name, FILL_BLANK(maxoptlen-optlen));
			if (CFG_tools[tool].options[i].size == 1)
				fprintf(stdout, "  %s%s", CFG_TypeNames[CFG_tools[tool].options[i].type], FILL_BLANK(8-typelen)); else
				fprintf(stdout, "  %s^%d%s", CFG_TypeNames[CFG_tools[tool].options[i].type], CFG_tools[tool].options[i].size, FILL_BLANK(8-typelen-2));
			fprintf(stdout, " %s%s  %s",
				CFG_tools[tool].options[i].default_val, FILL_BLANK(maxdeflen-deflen),
				CFG_tools[tool].options[i].desc);
			if (CFG_tools[tool].options[i].range == NULL || CFG_tools[tool].options[i].range[0] == 0) 
				fprintf(stdout, "\n"); else
				fprintf(stdout, " : %s\n", CFG_tools[tool].options[i].range);
		}
	}
	
	fprintf(stdout, "\n");
}

void CFG_printHelp(char *argv0, int tool)
{
	if (tool != -1) CFG_printToolHelp(argv0, tool); else
	{
		fprintf(stdout, "\n" clBGREEN "# TOOL usage" clNORMAL "\n");
		fprintf(stdout, "\n" clBYELLOW "  Usage:" clNORMAL "\n");
		fprintf(stdout, "    " clBWHITE "%s --usage\n" clNORMAL, argv0);
		fprintf(stdout, "    " clBWHITE "%s <obviously inconsistent arguments>\n" clNORMAL, argv0);
		
		fprintf(stdout, "\n" clBGREEN "# TOOL help" clNORMAL "\n");
		fprintf(stdout, "\n" clBYELLOW "  Usage:" clNORMAL "\n");
		fprintf(stdout, "    " clBWHITE "%s --help [", argv0);
		for(int i = 0; i < CFG_toolCount; i++) 
			fprintf(stdout, (i == 0 ? "%s" : " | %s"), CFG_tools[i].name);
		fprintf(stdout, "]" clNORMAL "\n");
		
		for(int i = 0; i < CFG_toolCount; i++)
			CFG_printToolHelp(argv0, i);
	}
}


#ifdef _USE_XML_
/**
init config
*/
void CFG_loadOptionsXML(char *xmlname, char *type, int tool)
{
	//--- puis on lit les options du fichier xml
	xmlDocPtr doc; xmlNodePtr cur;
	if (xmlname != NULL && xmlOpenFile(xmlname, &doc, &cur, type))
	{
		int i ;
		for( ; cur != NULL; cur = xmlGotoNext(cur))
		{
			for(i = 0; i < CFG_tools[tool].optionCount; i++)
			{
				if (CFG_tools[tool].options[i].type == pSep) continue;
				if (xmlKeyEq(cur, CFG_tools[tool].options[i].name)) 
				{ 
					char *val = xmlGetStrCont(doc, cur);
					CFG_tools[tool].options[i].val = 1;
					CFG_setVal(CFG_tools[tool].options+i, val); 
					xmlFree(val);
					break; 
				}
			}
			if (i == CFG_tools[tool].optionCount) WARN("Unknown xml option \"%s\"", cur->name);
		}
		xmlFreeDoc(doc);
	}
}
#endif

void CFG_loadOptionsCLI(int argc, char **argv)
{
	//--- et enfin on crase  nouveau par d'ventuelles options passes en paramtre
	int optlen;
	for (int j = 0; j < argc; j++)
		for(int i = 0; i < CFG_tools[CFG_tool].optionCount; i++)
		{
			if (CFG_tools[CFG_tool].options[i].type == pSep) continue;
			optlen = strlen(CFG_tools[CFG_tool].options[i].name); 
			if (argv[j][0] == '-' && !strncmp(argv[j]+1, CFG_tools[CFG_tool].options[i].name, optlen) && argv[j][optlen+1] == '=')
			{ 
				CFG_tools[CFG_tool].options[i].val = 2;
				CFG_setVal(CFG_tools[CFG_tool].options+i, argv[j]+optlen+2); 
				break; 
			}
			if (i == CFG_tools[CFG_tool].optionCount) WARN("Unknown CLI option \"%s\"", argv[i]);
		}
}

void CFG_printOptions()
{
	//--- avant de finalement afficher pour rappel toutes les valeurs effectives qui seront utilises
	fprintf(stdout, "\n" clBWHITE "Options:" clNORMAL "\n");
	for(int i = 0; i < CFG_tools[CFG_tool].optionCount; i++)
	{
		if (CFG_tools[CFG_tool].options[i].type == pSep) 
			fprintf(stdout, "\n"); 
		else
		{
			char markers[3][5] = { "[  ]", "[X ]", "[XX]" };
			int size = CFG_tools[CFG_tool].options[i].size;
			fprintf(stdout, "  %s -%s = " clBWHITE, markers[CFG_tools[CFG_tool].options[i].val], CFG_tools[CFG_tool].options[i].name);
			switch(CFG_tools[CFG_tool].options[i].type)
			{
				case pBool:
					if (size == 1) fprintf(stdout, "%d" clNORMAL "\n", *((bool*)(CFG_tools[CFG_tool].options[i].data)));
					else for(int j = 0; j < size; j++) fprintf(stdout, "%d%s", ((bool*)(CFG_tools[CFG_tool].options[i].data))[j], (j==size-1?clNORMAL "\n":" ; "));
					break;
				case pInt: case pEnum:
					if (size == 1) fprintf(stdout, "%d" clNORMAL "\n", *((int*)(CFG_tools[CFG_tool].options[i].data)));
					else for(int j = 0; j < size; j++) fprintf(stdout, "%d%s", ((int*)(CFG_tools[CFG_tool].options[i].data))[j], (j==size-1?clNORMAL "\n":" ; "));
					break;
				case pReal:
					if (size == 1) fprintf(stdout, "%g" clNORMAL "\n", *((real*)(CFG_tools[CFG_tool].options[i].data)));
					else for(int j = 0; j < size; j++) fprintf(stdout, "%g%s", ((real*)(CFG_tools[CFG_tool].options[i].data))[j], (j==size-1?clNORMAL "\n":" ; "));
					break;
				case pString: case pFile:
					if (size == 1) fprintf(stdout, "%s" clNORMAL "\n", *((char**)(CFG_tools[CFG_tool].options[i].data)));
					else for(int j = 0; j < size; j++) fprintf(stdout, "%s%s", ((char**)(CFG_tools[CFG_tool].options[i].data))[j], (j==size-1?clNORMAL "\n":" ; "));
					break;
				default: break;
			}
		}
	}
	fprintf(stdout, "\n");
}


#ifdef _USE_XML_
/**
write config xml
*/
void CFG_writeOptionsXml(char *type, FILE *file)
{
	fprintf(file, "<%s>\n", type);
	for(int i = 0; i < CFG_tools[CFG_tool].optionCount; i++)
	{
		if (CFG_tools[CFG_tool].options[i].type == pSep) 
			fprintf(file, "\n"); 
		else
		{
			switch(CFG_tools[CFG_tool].options[i].type)
			{
				case pBool:
					fprintf(file, "  <%s>%d</%s>\n", CFG_tools[CFG_tool].options[i].name, *((bool*)(CFG_tools[CFG_tool].options[i].data)), CFG_tools[CFG_tool].options[i].name);
					break;
				case pInt: case pEnum:
					fprintf(file, "  <%s>%d</%s>\n", CFG_tools[CFG_tool].options[i].name, *((int*)(CFG_tools[CFG_tool].options[i].data)), CFG_tools[CFG_tool].options[i].name);
					break;
				case pReal:
					fprintf(file, "  <%s>%g</%s>\n", CFG_tools[CFG_tool].options[i].name, *((real*)(CFG_tools[CFG_tool].options[i].data)), CFG_tools[CFG_tool].options[i].name);
					break;
				case pString:
					fprintf(file, "  <%s>%s</%s>\n", CFG_tools[CFG_tool].options[i].name, *((char**)(CFG_tools[CFG_tool].options[i].data)), CFG_tools[CFG_tool].options[i].name);
					break;
				case pFile: // here we take str and not data because of the possible path correction
					fprintf(file, "  <%s>%s</%s>\n", CFG_tools[CFG_tool].options[i].name, CFG_tools[CFG_tool].options[i].str, CFG_tools[CFG_tool].options[i].name);
					break;				
				default: break;
			}
		}
	}
	fprintf(file, "</%s>\n", type);
}
#endif


// intermediate functions

void CFG_setVal(CFG_Param *param, char *arg)
{
	delete_(param->str); param->str = strclone(arg);
	switch(param->type)
	{
		case pBool:
			if (param->size == 1) 
			{
				if (strstr(arg, "true") || strstr(arg, "yes")) *((bool*)(param->data)) = true; else
				if (strstr(arg, "false") || strstr(arg, "no")) *((bool*)(param->data)) = false; else
				*((bool*)(param->data)) = strtoi(arg);
			}
			else
			{
				char *arg1 = arg, *arg2 = arg;
				bool cut;
				for(int i = 0; i < param->size; i++) 
				{
					for(arg2 = arg1; *arg2 != 0 && *arg2 != sep; arg2++);
					if (*arg2 == sep) { *arg2 = 0; cut = true; } else cut = false;
					if (strstr(arg1, "true") || strstr(arg1, "yes")) (*((bool**)(param->data)))[i] = true; else
					if (strstr(arg1, "false") || strstr(arg1, "no")) (*((bool**)(param->data)))[i] = false; else
					(((bool*)(param->data)))[i] = strtoi(arg1);
					if (cut) *arg2 = sep;
					if (*arg2 == 0 && i < param->size-1)// ERROR("missing values in %s (expected %d)", arg, param->size);
					{
						for(int j = i+1; j < param->size; j++) (((bool*)(param->data)))[j] = (((bool*)(param->data)))[i];
						break; 
					}
					arg1 = arg2+1;
				}
			}
			break;
	
		case pInt: case pEnum:
			if (param->size == 1) 
				*((int*)(param->data)) = strtoi(arg);
			else
			{
				char *arg1 = arg, *arg2 = arg;
				bool cut;
				for(int i = 0; i < param->size; i++) 
				{
					for(arg2 = arg1; *arg2 != 0 && *arg2 != sep; arg2++);
					if (*arg2 == sep) { *arg2 = 0; cut = true; } else cut = false;
					(((int*)(param->data)))[i] = strtoi(arg1);
					if (cut) *arg2 = sep;
					if (*arg2 == 0 && i < param->size-1) //ERROR("missing values in %s (expected %d)", arg, param->size);
					{
						for(int j = i+1; j < param->size; j++) (((int*)(param->data)))[j] = (((int*)(param->data)))[i];
						break; 
					}
					arg1 = arg2+1;
				}
			}
			break;
	
		case pReal:
			if (param->size == 1) 
				*((real*)(param->data)) = atof(arg);
			else
			{
				char *arg1 = arg, *arg2 = arg;
				bool cut;
				for(int i = 0; i < param->size; i++) 
				{
					for(arg2 = arg1; *arg2 != 0 && *arg2 != sep; arg2++);
					if (*arg2 == sep) { *arg2 = 0; cut = true; } else cut = false;
					(((real*)(param->data)))[i] = atof(arg1);
					if (cut) *arg2 = sep;
					if (*arg2 == 0 && i < param->size-1) //ERROR("missing values in %s (expected %d)", arg, param->size);
					{
						for(int j = i+1; j < param->size; j++) (((real*)(param->data)))[j] = (((real*)(param->data)))[i];
						break; 
					}
					arg1 = arg2+1;
				}
			}
			break;
		
		case pString: case pFile:
			if (param->size == 1)
			{
				if (param->type == pFile && param->val == 1 && arg)
					*((char**)(param->data)) = pathcat(CFG_configFile, arg); else
					*((char**)(param->data)) = strclone(arg);
			}
			else
			{
				char *arg1 = arg, *arg2 = arg;
				bool cut;
				for(int i = 0; i < param->size; i++) 
				{
					for(arg2 = arg1; *arg2 != 0 && *arg2 != sep; arg2++);
					if (*arg2 == sep) { *arg2 = 0; cut = true; } else cut = false;
					if (param->type == pFile && param->val == 1)
						(((char**)(param->data)))[i] = pathcat(CFG_configFile, arg1); else
						(((char**)(param->data)))[i] = strclone(arg1);
					if (cut) *arg2 = sep;
					if (*arg2 == 0 && i < param->size-1) //{ ERROR("missing values in %s (expected %d)", arg, param->size); break; }
					{
						for(int j = i+1; j < param->size; j++) (((char**)(param->data)))[j] = (((char**)(param->data)))[i];
						break; 
					}
					arg1 = arg2+1;
				}
			}
			break;
			
		default: break;
	}
}

