/********************************************
Personal library of Cyril Roussillon
http://crteknologies.free.fr
cyril42e@gmail.com
2006-2007
*********************************************/

#ifndef CYRIL_TIMER_H
#define CYRIL_TIMER_H

#ifdef __cplusplus
	extern "C" {
#endif


//////////////////////////////////////////////////////////////////
// TimeStamp
//////////////////////////////////////////////////////////////////
// void ____________() {}

#include <sys/time.h>
#include <sys/times.h>
#include <time.h>
#include "cyril-macro.h"

typedef struct TimeStamp_
{
	struct timeval tv;
	struct timezone tz;
	struct tms ts;
	Date startingValue;
} TimeStamp;

extern TimeStamp ts;




/*
// get calendar seconds
extern Date get_c_s() ;
// get calendar milliseconds
extern Date get_c_ms();
// get calendar microseconds
extern Date get_c_us();
// get user cpu clocks
extern Date get_u_clk();
// get user cpu microseconds
extern Date get_u_us();

// get system cpu clocks
extern Date get_s_clk();
// get system cpu microseconds
extern Date get_s_us();


// get user and system cpu clocks
extern Date get_us_clk();
// get user and system cpu microseconds
extern Date get_us_us();
*/

typedef struct Timer_
{
//	int vals;
	Date calendar;
	Date user_cpu;
	Date system_cpu;
} Timer;

typedef struct Chrono_
{
	Timer timer;
	Timer timer_tot;
	int vals;
} Chrono;



// get calendar seconds
extern inline Date get_c_s() 
{
	gettimeofday(&ts.tv,&ts.tz);
	return (ts.tv.tv_sec - ts.startingValue);
}
// get calendar milliseconds
extern inline Date get_c_ms() 
{
	gettimeofday(&ts.tv,&ts.tz);
	return (ts.tv.tv_sec - ts.startingValue)*1000 + ts.tv.tv_usec/1000;
}
// get calendar microseconds
extern inline Date get_c_us() 
{
	gettimeofday(&ts.tv,&ts.tz);
	return (ts.tv.tv_sec - ts.startingValue)*1000000 + ts.tv.tv_usec;
}

// get user cpu clocks
extern inline Date get_u_clk()
{
	times(&ts.ts);
	return ts.ts.tms_utime;
}
// get user cpu microseconds
extern inline Date get_u_us()
{
	return (Date)(get_u_clk()*1.E6 / CLOCKS_PER_SEC);
}

// get system cpu clocks
extern inline Date get_s_clk() 
{
	times(&ts.ts);
	return ts.ts.tms_stime;
}
// get system cpu microseconds
extern inline Date get_s_us()
{
	return (Date)(get_s_clk()*1.E6 / CLOCKS_PER_SEC);
}


// get user and system cpu clocks
extern inline Date get_us_clk()
{
//	times(&ts.ts);
//	return ts.ts.tms_stime + ts.ts.tms_utime;
	return times(&ts.ts);
}
// get user and system cpu microseconds
extern inline Date get_us_us() //;
{
	return (Date)(get_us_clk()*1.E6 / CLOCKS_PER_SEC);
}



extern inline void timer_start(Timer *c)
{
	times(&ts.ts);
	gettimeofday(&ts.tv,&ts.tz);
	c->calendar = (ts.tv.tv_sec - ts.startingValue)*1000000 + ts.tv.tv_usec;
//	c->user_cpu = (Date)(ts.ts.tms_utime*1.E6 / CLOCKS_PER_SEC);
//	c->system_cpu = (Date)(ts.ts.tms_stime*1.E6 / CLOCKS_PER_SEC);
	c->user_cpu = (Date)ts.ts.tms_utime;
	c->system_cpu = (Date)ts.ts.tms_stime;
}

extern inline void timer_stop(Timer *c)
{
	gettimeofday(&ts.tv,&ts.tz);
	times(&ts.ts);
	c->calendar = ((ts.tv.tv_sec - ts.startingValue)*1000000 + ts.tv.tv_usec) - c->calendar;
//	c->user_cpu = (Date)(ts.ts.tms_utime*1.E6 / CLOCKS_PER_SEC) - c->user_cpu;
//	c->system_cpu = (Date)(ts.ts.tms_stime*1.E6 / CLOCKS_PER_SEC) - c->system_cpu;
	c->user_cpu = (Date)ts.ts.tms_utime - c->user_cpu;
	c->system_cpu = (Date)ts.ts.tms_stime - c->system_cpu;
}

extern inline void timer_add(Timer *moy, Timer *c)
{
	moy->calendar += c->calendar;
	moy->user_cpu += c->user_cpu;
	moy->system_cpu += c->system_cpu;
}
/*
void timer_best(Timer *best, Timer *c)
{
	if (best->vals == 1)
	{
		if (c->calendar < best->calendar) best->calendar = c->calendar;
		if (c->user_cpu < best->user_cpu) best->user_cpu = c->user_cpu;
		if (c->system_cpu < best->system_cpu) best->system_cpu = c->system_cpu;
	} else
	{
		best->calendar = c->calendar;
		best->user_cpu = c->user_cpu;
		best->system_cpu = c->system_cpu;
		best->vals = 1;
	}
}
*/


extern inline void chrono_start(Chrono *c)
{
	timer_start(&c->timer);
}
extern inline void chrono_stop(Chrono *c)
{
	timer_stop(&c->timer);
	timer_add(&c->timer_tot, &c->timer);
}
extern inline void chrono_reset(Chrono *c)
{
	c->timer_tot = (Timer){0,0,0};
	c->vals = 0;
}

extern inline void chrono_norm(Chrono *c)
{
	c->timer_tot.calendar /= c->vals;
	c->timer_tot.user_cpu /= c->vals;
	c->timer_tot.system_cpu /= c->vals;
	c->vals = 1;
}

extern inline void chrono_print(Chrono *c)
{
	fprintf(stdout, "wc %lld, user %lld, sys %lld\n", 
		c->timer_tot.calendar, c->timer_tot.user_cpu, c->timer_tot.system_cpu);
}


/*
extern void timer_start(Timer *c);
extern void timer_stop(Timer *c);
extern void timer_add(Timer *moy, Timer *c);

extern void chrono_start(Chrono *c);
extern void chrono_stop(Chrono *c);
extern void chrono_reset(Chrono *c);
extern void chrono_norm(Chrono *c);
extern void chrono_print(Chrono *c);
*/
/*
extern void chrono_start(Chrono *c);
extern void chrono_stop(Chrono *c);
extern void chrono_print(Chrono *c);
extern void chrono_init(Chrono *c);

extern void chrono_best(Chrono *best, Chrono *c);
extern void chrono_sum(Chrono *moy, Chrono *c);
extern void chrono_norm(Chrono *c);
*/

#ifdef __cplusplus
	}
#endif

#endif
