
#include "config.h"
#include "cyril-config.h"
#include "vfh.h"
#include "astar.h"
#include "io.h"
#include "hclient.h"
#include "cyril-timer.h"

#include <list>

#define ROBOT_

int main(int argc, char **argv)
{
	load_config();
	CFG_loadArgs(argc, argv);
	if (CFG_tool < 0) return 0;
	CFG_printOptions();
	
	/// init
	Robot robot(33, 28, 8);
	for(int i = 0; i < robot.nsensors; i++)
		robot.readings[i].is_working = true;
//	robot.readings[4].is_working = true;
	
	robot.state.x = config.start_x;
	robot.state.y = config.start_y;
	robot.state.theta = config.start_theta;
	robot.state.timestamp = get_c_us();
	robot.minSensorRange = 200;
	robot.maxSensorRange = 500;

	GlobalMap gm(config.global_map_file, config.cell_size);
	VFH vfh(&robot, gm.w, gm.h);

	/// running
	bool goal_reached = false;
	int inter_goal_x = (int)(config.goal_x/gm.cellSize);
	int inter_goal_y = (int)(config.goal_y/gm.cellSize);
	gm.add_goal_graph(inter_goal_x, inter_goal_y);
	bool stop_at_goal = !config.use_planning;

//	Date control_timer = 0;
	Date planning_timer = 0;
	Date recordmaps_timer = 0;
	Date tmp_timer;
	int map_number = 0;
	char map_filename[100];
	std::list<Vertex *> list_inter; list_inter.push_back(gm.goal_vtx);

//list_inter = findpath_opt1(gm, robot.state.x, robot.state.y, gm.goal_vtx);
//DEBUG("size %d", list_inter.size());

	
//	vfh.saveMaps("map_test.ppm", gm, &list_inter, 90);
//	return 0;

#ifdef ROBOT_	
	robot.init("fred");
#endif

/*	
	hclient_drive(0.3);
	sleep(2);
	hclient_drive(0);
	hclient_steer(30);
	sleep(4);
	hclient_steer(30);
	sleep(4);
	hclient_steer(-30);
	sleep(4);
	hclient_steer(0);
	robot.close();
	return 0;
	*/

/*	robot.setDirSpeed(45, 20);
	sleep(2);
	robot.setDirSpeed(45, 20);
	sleep(2);
	robot.setDirSpeed(45, 20);
	sleep(2);
	robot.setDirSpeed(45, 20);
	sleep(2);
	robot.setDirSpeed(45, 20);
	sleep(2);
	robot.setDirSpeed(45, 20);
	sleep(2);
	robot.close();
	return 0;
*/
	
	
	while (!goal_reached)
	{
		// planning
		if (config.use_planning && ((tmp_timer = get_c_ms()) >= planning_timer+config.planning_period))
		{
DEBUG("plan!");
			planning_timer = tmp_timer;
			list_inter = findpath_opt1(gm, robot.state.x, robot.state.y, gm.goal_vtx);
			if (list_inter.size() == 0) { ERROR("no path possible !!"); return 0; }
			if (list_inter.size() == 1) stop_at_goal = true;
			Vertex *intermediary_goal = list_inter.front();
			inter_goal_x = intermediary_goal->x;
			inter_goal_y = intermediary_goal->y;
		}
	
		// sensing
#ifdef ROBOT_	
		vfh.robot->updateReadings();
		vfh.updateMap();
#endif
		
		//control
		float speed, dir;
		vfh.getDirSpeed(dir, speed, inter_goal_x*gm.cellSize, inter_goal_y*gm.cellSize, stop_at_goal);
#ifdef ROBOT_	
		robot.setDirSpeed(dir, speed);
#endif

		if (stop_at_goal && (sqr(robot.state.x-inter_goal_x*gm.cellSize)+sqr(robot.state.y-inter_goal_y*gm.cellSize)) < 200)
			goal_reached = true;

DEBUG("loop (map %d): pos %f,%f,%f ; goal is %f,%f, stop %d, reached %d ; control t %f v %f",
	map_number, robot.state.x, robot.state.y, robot.state.theta,
	inter_goal_x*gm.cellSize, inter_goal_y*gm.cellSize, stop_at_goal?1:0, goal_reached?1:0, dir, speed);
			
		// save maps
		if (config.record_maps && ((tmp_timer = get_c_ms()) >= recordmaps_timer+160))
		{
			sprintf(map_filename, "maps/map%d_%04d.ppm", getpid(), map_number++);
			vfh.saveMaps(map_filename, gm, &list_inter, dir);
		}
		
		// check robot still inside map
		if (robot.state.x < 40 || robot.state.y < 40 || 
			robot.state.x > gm.w*gm.cellSize-40 || robot.state.y > gm.h*gm.cellSize-40)
		{
			ERROR("robots got out of the map");
			break;
		}
			

	}

	robot.close();
	
	return 0;
}


/*	robot.sensors[0] = Sensor(6, -2, 16, 12); // x,y,theta
	robot.sensors[1] = Sensor(7, -8, 11, 44);
	robot.sensors[2] = Sensor(0, -12, 7, 90);
	robot.sensors[3] = Sensor(1, -5, -14, 144);
	robot.sensors[4] = Sensor(2, 5, -14, -144);
	robot.sensors[5] = Sensor(3, 12, 7, -90);
	robot.sensors[6] = Sensor(4, 8, 11, -44);
	robot.sensors[7] = Sensor(5, 2, 16, -12);
*/


