#ifndef _VFH_H_
#define _VFH_H_

#include "datastruct.h"
#include "io.h"


/***************************************
LocalMap
***************************************/

struct CertaintyCell
{
	int probed; /// how many times the element has been probed
	int obst; /// how many times the element has been tested as an obstacle
	CertaintyCell() { probed = 0; obst = 0; }
	CertaintyCell(int probed, int obst): probed(probed), obst(obst) {}
};

class LocalMap
{
 public:
	// state
	int w, h; /// size of global map
	int size; /// size of local map
	int x, y; /// position of local map inside global map
	CertaintyCell **map; /// size w*h
	float cellSize;
	// constructors
	LocalMap(int w, int h, int size, float cellSize);
	~LocalMap();
	// methods
	void shift(int x, int y, bool memory);
};

/***************************************
PolarMap
***************************************/

class PolarMap
{
 public:
	// state
	int size;
	float *map; /// size size
	float *tmp_map;
	// constructors
	PolarMap(int size);
	~PolarMap();
	// methods
	void compute(LocalMap &lm, RobotState &rs);
	float getDirection(RobotState &rs, float xGoal, float yGoal, float histo_thres);
	float getSpeed(RobotState &rs, float xGoal, float yGoal, bool stop_at_goal, float speed_reduction);
};


/***************************************
VFH
***************************************/

class VFH
{
 public:
	// state
	LocalMap *lm;
	PolarMap *pm;
	Robot *robot;
	float xGoal, yGoal;
	bool stop_at_goal;
	uchar *image;
	float histo_thres, speed_reduction;
	// constructors
	VFH(Robot *robot, int mapw, int maph);
	~VFH();
	// methods
	void updateMap();
	void getDirSpeed(float &theta, float &vel, float goal_x, float goal_y, bool stop_at_goal);
	void saveMaps(char *filename, GlobalMap &gm, std::list<Vertex *> *list_inter, float direction);
	
};




#endif
